/* 

    umesh2dgui, a GUI for umesh2d grid generator 

    Copyright (C) 2003 Gabriele Dini Ciacci
  
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include "u2d_core.h"

bool isnear(double toll, double x, double y){

  if(fabs(x - y)  > toll) return false;
  else return true;

}

bool isnear(double toll, double x1, double y1, double x2, double y2){

  if(hypot(x1-x2,y1-y2) > toll) return false;
  else return true;

}

bool isnear(double toll, Primitive *primX, Primitive *primY){
  /* for now do not use this */
  /* we must check 4 combinations of points and first I have to
   * figure out witch is better way to do this*/

  //  rvalue = false;
  //  if(isnear(toll, primX->x1, primY->y1, primX->x2, primY->y2))
  return false;

}

Primitive::Primitive(){
  nextprimitive = NULL;
  prevprimitive = NULL;
  maxdata = 0;
  curdata = 0;
  data = NULL;
  object = 0;
  object_part = 0;
  isboundary = false;
  beginvertex = 0;
  endvertex = 0;
}


double Primitive::getdata(int x, int y) {
  double localdata;
  localdata = this->data[y+3*x];
  return localdata;
}

void Primitive::setdata(int x, int y, double value) {
  this->data[y+3*x] = value;
}

int Primitive::allocatedata(int ndata) {
  if(ndata == 0){
    /* We requested to allocate zero data */
    fprintf(stdout, "Requested an alocation of zero [i|s|h]data, skipping data allocation\n");
    return 0;
  }
  else if(this->data == NULL) {
    if(this->maxdata == 0) {
      fprintf(stdout, "Allocating a new mem chunk for [i|s|h]data for " \
                      "%i elements (%i bytes): ", ndata, ndata*3*sizeof(double));
      this->data = (double *)calloc(sizeof(double), ndata * 3);
      if(this->data == NULL) { 
	fprintf(stdout, "Failed\n");
	fprintf(stderr, "[i|s|h]data allocation failed\n");
	return -1;  // memory allocation failed
      }
      else {
	fprintf(stdout, "Success\n");
	this->maxdata = ndata;
	return 0; // OK!
      }
    }
    else {
      fprintf(stderr, "WTF?! data is null but maxdata is not 0!!\n");
      return -2; // Primitive data mismatch
    }
  }
  else {
    if(ndata < (this->maxdata) * 2) { 
      /* We dont' reallocate memory if the mem avaiable is allready 
       * big enougth but less than double of avaiable. We waste but 
       * not too much
       */ 
      fprintf(stdout, "Data chunk was allready enougth big (but less" \
                      "than double), we waste mem but save cpu\n");
      return 0;
    }
    else {
      fprintf(stdout, "Reallocating old mem chunk of %i elements (%i bytes)" \
                      "to a new chunk of %i elements (%i bytes): ", \
                       this->maxdata, this->maxdata*3*sizeof(double),\
                       ndata, ndata*3*sizeof(double));
      this->data = (double *)realloc(this->data, 3*sizeof(double)*ndata);
      if(this->data == NULL) {
	fprintf(stdout, "Failed\n");
	fprintf(stderr, "[i|s|h]data reallocation failed\n");
	return -1;  // memory reallocation failed
      }
      else {
	fprintf(stdout, "Success\n");
	this->maxdata = ndata;
	return 0; // OK!
      }
    }
  }
  return -3; // How the hell we got here???!!!!
}

void Line::set_coord(int spli,\
		     double X1, double Y1,\
		     double X2, double Y2){
  spline = spli;
  x1 = X1;
  y1 = Y1;
  x2 = X2;
  y2 = Y2;
}

void Circle::find_extremes(){
  x1 = centerx + radius*cos(startarc / 360 * 2.0 * PI);
  y1 = centery + radius*sin(startarc / 360 * 2.0 * PI);
  x2 = centerx + radius*cos(endarc / 360 * 2.0 * PI);
  y2 = centery + radius*sin(endarc / 360 * 2.0 * PI);
}

void Circle::set_coord(int spli, double rad,\
		       double centx, double centy,\
		       double sarc, double earc){
  this->spline = spli;
  this->radius = rad;
  this->centerx = centx;
  this->centery = centy;
  this->startarc = sarc;
  this->endarc = earc;
  this->find_extremes();
}

void Ellipse::find_extremes(){
  x1 = centerx + semix * cos(startarc / 360 * 2.0 * PI);
  y1 = centery + semiy * sin(startarc / 360 * 2.0 * PI);
  x2 = centerx + semix * cos(endarc / 360 * 2.0 * PI);
  y2 = centery + semiy * sin(endarc / 360 * 2.0 * PI);
}

void Ellipse::set_coord(int spli,\
			double semx, double semy,\
			double centx, double centy,\
			double sarc, double earc){
  this->spline = spli;
  this->semix = semx;
  this->semiy = semy;
  this->centerx = centx;
  this->centery = centy;
  this->startarc = sarc;
  this->endarc = earc;
  this->find_extremes();
}


u2d_Core::u2d_Core(){
  firstprimitive = NULL;
  lastprimitive = NULL;
  globalprimid = 0;
  elements = 0;
  memset(prim_string, 0, 1024);
  vertex_n = 1;
  last_object = 0;
}


/* PRIMITIVE LIST MANGEMENT */

Primitive *u2d_Core::new_prim(int type) {
  Primitive *prim = NULL;
  elements++;
  //  vertexs++;
  switch(type) {
  case PR_LINE:
    prim = new Line;
    break;
  case PR_CIRCLE:
    prim = new Circle;
    break;
  case PR_ELLIPSE:
    prim = new Ellipse;
    break;
  case PR_DATA:
    prim = new Data;
    break;
  default:
    elements--;
    //    vertexs--;
    fprintf(stderr, "Unknown primitive type\n");
  }
  return prim;
}

Primitive *u2d_Core::add_prim(int type) {
  Primitive *newprim;
  newprim = new_prim(type);
    if(firstprimitive == NULL){
      firstprimitive = newprim;
    }
    else {
      lastprimitive->nextprimitive = newprim;
    }
    newprim->prevprimitive = lastprimitive;
    lastprimitive = newprim;
    newprim->id = globalprimid++;
    return newprim;
}

void u2d_Core::del_prim(Primitive *oldprim) {
  if(oldprim == NULL){
    /* Just to be sure none pass us a NULL pointer */
    return;
  }
  else if(oldprim == firstprimitive && oldprim->nextprimitive != NULL){
    /* We want to delete the firstprimitive and it's not the only one */
    firstprimitive = oldprim->nextprimitive;
    oldprim->nextprimitive->prevprimitive = NULL;
  }
  else if(oldprim == firstprimitive && oldprim->nextprimitive == NULL){
    /* We want to delete the firstprimitive and it's the only one */
    firstprimitive = NULL;
    lastprimitive = NULL;
  }
  else if(oldprim->nextprimitive != NULL) {
    /* oldprim is not the first, because we checked that above
     * and is not the last because it points to anoter one. It
     * is in the middle of the list */
    oldprim->prevprimitive->nextprimitive = oldprim->nextprimitive;
    oldprim->nextprimitive->prevprimitive = oldprim->prevprimitive;
  }
  else if (oldprim->nextprimitive == NULL) {
    /* oldprim should be the last, but it is not the first and only
     * because we checked that above */
    oldprim->prevprimitive->nextprimitive = NULL;
    lastprimitive = oldprim->prevprimitive;
  }
  delete oldprim;
  elements--;
}

int u2d_Core::get_prim_number(){
  return elements;
}

Primitive *u2d_Core::get_prim_list_start(){
  return firstprimitive;
}
Primitive *u2d_Core::get_prim_list_end(){
  return lastprimitive;
}

/* I could have impemented a get_prim_string method
 * for each Primitive class type, but for speed and
 * space list 8we have many istances of Primitive
 * class) i have preferred to use an old "static"
 * function placed in the core class. It is real faster
 */
char *u2d_Core::get_prim_string(Primitive *curprim) {
  memset(prim_string, 0, 1024);

  /* Changed to the more C style below
  char primname[34];
  if(strlen(curprim->primname) > 0){
    strncpy(primname, curprim->primname, 34);
  }
  else{
    strncpy(primname, "None", 34);
  }
  */
  switch(curprim->type) {
  case PR_LINE:
    snprintf(prim_string, 1024, \
	     "%i\t%s\t%i\tLine\t%0.2f\t%0.2f\t%0.2f\t%0.2f",
	     curprim->id,
	     strlen(curprim->primname) > 0 ? curprim->primname : "None",
	     curprim->object,
	     ((Line *)curprim)->x1,
	     ((Line *)curprim)->y1,
	     ((Line *)curprim)->x2,
	     ((Line *)curprim)->y2
	     );
    break;
  case PR_CIRCLE:
    snprintf(prim_string, 1024, \
	     "%i\t%s\t%i\tCircle\t%0.2f\t%0.2f\t\t\t%0.2f\t%0.2f\t%0.2f",
	     curprim->id,
	     strlen(curprim->primname) > 0 ? curprim->primname : "None",
	     curprim->object,
	     ((Circle *)curprim)->centerx,
	     ((Circle *)curprim)->centery,
	     ((Circle *)curprim)->radius,
	     ((Circle *)curprim)->startarc,
	     ((Circle *)curprim)->endarc
	     );
    break;
  case PR_ELLIPSE:
    snprintf(prim_string, 1024, \
	     "%i\t%s\t%i\tEllipse\t%0.2f\t%0.2f\t\t\t%0.2f\t%0.2f\t%0.2f",
	     curprim->id,
	     strlen(curprim->primname) > 0 ? curprim->primname : "None",
	     curprim->object,
	     ((Ellipse *)curprim)->centerx,
	     ((Ellipse *)curprim)->centery,
	     ((Ellipse *)curprim)->semix,
	     ((Ellipse *)curprim)->startarc,
	     ((Ellipse *)curprim)->endarc
	     );
    break;
  case PR_DATA:
    /* to implement */
    break;
  default:
    memset(prim_string, 0, MAX_PRIM_STRING);
    break;
  }
  return prim_string;
}

Primitive *u2d_Core::get_prim_by_label(const char *label) {
  int id = atoi(label);
  Primitive *curprim = firstprimitive;
  while(curprim){
    if(curprim->id == id){
      return curprim;
    }
    else {
      curprim = curprim->nextprimitive;
    }
  }
  fprintf(stderr, "Primitive not found! Browser list and primitive " \
	  "list missmatch! Or id %i unexistant!!\n", id);
  return NULL;
}

int u2d_Core::get_vertex_n(){
  return vertex_n;
}

void u2d_Core::find_objects() {
  if(firstprimitive == NULL) {
    /* Primitive list is still empty */
    return;
  }
  /* Here we reset all the variables used by calculation:
   * * object in each primitive
   * * object_part in each primitive (progressive object number)
   * * last_object global variable
   */
  last_object = 0;
  bool attached = false;
  Primitive *curprim = firstprimitive;
  vertex_n = 0;
  while(curprim){
    curprim->object = 0;
    curprim->object_part = 0;
    curprim->beginvertex = 0;
    curprim->endvertex = 0;
    curprim = curprim->nextprimitive;
  }


  /* THE algo */
  curprim = firstprimitive;
  Primitive *otherprim = firstprimitive;
  Primitive *firstprim = NULL;
  Primitive *lastprim = NULL;
  while(curprim){
    if(!curprim->object) {
      /* We do not flag the current primitive as used, so it stay in
       * the comparison pool of primitives. This way when we are the
       * last primitive and we are looking back for the first (the
       * one from witch the object start) we can find it among
       * those that are unflagged.
       */
      /* curprim->object = ++last_object; */
      last_object++;
      curprim->object_part++;
      attached = false;
      otherprim = firstprimitive;
      firstprim = curprim;
      lastprim = curprim;
      while(otherprim){
	attached = false;
	/* Primitive is allready taken or is the curprim itself*/
	if(otherprim->object || (otherprim == curprim)) {
	  otherprim = otherprim->nextprimitive;
	}
	else {

	  if(curprim->beginvertex == 0 && \
	     isnear(DEFAULT_TOLL, curprim->x1, curprim->y1, otherprim->x1, otherprim->y1)){
	    vertex_n++;
	    curprim->beginvertex = vertex_n;
	    otherprim->beginvertex = vertex_n;
	    attached = true;
	  }
	  else if(curprim->beginvertex == 0 && \
		  isnear(DEFAULT_TOLL, curprim->x1, curprim->y1, otherprim->x2, otherprim->y2)){
	    vertex_n++;
	    curprim->beginvertex = vertex_n;
	    otherprim->endvertex = vertex_n;
	    attached = true;
	  }

	  if(curprim->endvertex == 0 && \
	     isnear(DEFAULT_TOLL, curprim->x2, curprim->y2, otherprim->x1, otherprim->y1)){
	    vertex_n++;
	    curprim->endvertex = vertex_n;
	    otherprim->beginvertex = vertex_n;
	    attached = true;
	  }
	  else if(curprim->endvertex == 0 && \
		  isnear(DEFAULT_TOLL, curprim->x2, curprim->y2, otherprim->x2, otherprim->y2)){
	    vertex_n++;
	    curprim->endvertex = vertex_n;
	    otherprim->endvertex = vertex_n;
	    attached = true;
	  }

	  if(attached == true){
	    /* Primitive otherprim is attached */
	    otherprim->object = last_object;
	    otherprim->object_part = ++curprim->object_part;
	    /* We check that the current primitive is really the last
	     * before assigning it to lastprim, since maybe we are
	     * just getting the firstprim again from the pool
	     * to close the object. Remember from above note that
	     * we did not flagged the firstprimitive of the object
	     * as used
	     */
	    if(otherprim != firstprim){
	      lastprim = otherprim;
	    }
	    curprim = otherprim;
	    otherprim = firstprimitive;
	  }
	  else {
	    /* Primitive otherprim is not attached */
	    otherprim = otherprim->nextprimitive;
	  }
	}


	/* firstprim and lastprim should be the first and
	 * last primitive of the object
	 */
	if(otherprim == NULL){
	  /* We passed the entire list without finding
	   * anymore primitives to add to this object
	   */
	  if(firstprim == lastprim){
#ifdef DEBUGGING
	    printf("Primitive %s is closed on itself\n", get_prim_string(firstprim));
#endif
	    firstprim->beginvertex = ++vertex_n;
	    firstprim->endvertex = vertex_n;
	    curprim->object = ++last_object;
	  }
	}

      }
    }
    curprim = curprim->nextprimitive;
  }


}

